/****************************************************************************
 * rt2400 Linux Driver
 * http://rt2400.sourceforge.net
 * Refer to the LICENSE file for code license
 *
 * Parts of this source were from original RaLink drivers.
 ****************************************************************************

	Module Name:
	rtmp_main.c

	Abstract:

	Revision History:
	Who		When			What
	--------	----------		-----------------------------
	Paul Lin	25-11-2002		Initial version
	MarkW           09-02-2004		Baseline of code
	defekt          13-02-2004              ATOMIC kmalloc fix
	MarkW		13-03-2004		EXPORT_NO_SYMBOLS 2.6 fix
	
*/

#include "rt_config.h"

//	Global static variable, Debug level flag
#ifdef RT2400_DBG
ULONG	RTDebugLevel = RT_DEBUG_TRACE;
#endif

// The driver version information
static char version[] __devinitdata = DRV_NAME " " DRV_VERSION " " DRV_RELDATE " Paul Lin and others\n";

// Following information will be show when you run 'modinfo'
MODULE_AUTHOR("http://rt2400.sourceforge.net");
MODULE_DESCRIPTION("Ralink RT2400 802.11b WLAN driver " DRV_VERSION " " DRV_RELDATE);
#if LINUX_VERSION_CODE >= 0x20412		// Red Hat 7.3
MODULE_LICENSE("GPL");
#endif

static INT __devinit RT2400_init_one (
	IN	struct pci_dev				*pPci_Dev,
	IN	const struct pci_device_id	*ent)
{
	INT rc;

	// wake up and enable device
	if (pci_enable_device (pPci_Dev))
	{
		rc = -EIO;
	}
	else
	{
		rc = RT2400_probe(pPci_Dev, ent);		
	}
	return rc;
}

#if LINUX_VERSION_CODE <= 0x20402		// Red Hat 7.1
static struct net_device *alloc_netdev(int sizeof_priv, const char *mask, void (*setup)(struct net_device *))
{
	struct net_device *dev;
	int alloc_size;

	/* ensure 32-byte alignment of the private area */
	alloc_size = sizeof (*dev) + sizeof_priv + 31;

	dev = (struct net_device *) kmalloc (alloc_size, GFP_ATOMIC);
	if (dev == NULL)
	{
		printk(KERN_ERR "alloc_netdev: Unable to allocate device memory.\n");
		return NULL;
	}

	memset(dev, 0, alloc_size);

	if (sizeof_priv)
		dev->priv = (void *) (((long)(dev + 1) + 31) & ~31);
		
	setup(dev);
	strcpy(dev->name,mask);

	return dev;
}
#endif

//
// PCI device probe & initialization function
//
INT __devinit   RT2400_probe(
	IN	struct pci_dev			*pPci_Dev, 
	IN	const struct pci_device_id	*ent)
{
	struct	net_device		*net_dev;
	RTMP_ADAPTER			*pAd;
	CHAR					*print_name;
	INT						chip_id = (int) ent->driver_data;
	ULONG					csr_addr;
	CSR3_STRUC				StaMacReg0;
	CSR4_STRUC				StaMacReg1;
	INT						Status;
	
	DBGPRINT(RT_DEBUG_TRACE, "%s\n",version);

	print_name = pPci_Dev ? pPci_Dev->slot_name : "rt2400";

#if LINUX_VERSION_CODE <= 0x20402		// Red Hat 7.1
	net_dev = alloc_netdev(sizeof(RTMP_ADAPTER), "eth%d", ether_setup);
#else
	net_dev = alloc_etherdev(sizeof(RTMP_ADAPTER));
#endif
	if (net_dev == NULL) 
	{
		DBGPRINT(RT_DEBUG_TRACE, "init_ethernet failed\n");
		goto err_out;
	}
	
	SET_MODULE_OWNER(net_dev);
		
	if (pci_request_regions(pPci_Dev, print_name))
		goto err_out_free_netdev;
	
	// Interrupt IRQ number
	net_dev->irq = pPci_Dev->irq;
	
	// map physical address to virtual address for accessing register
	csr_addr = (ULONG) ioremap(pci_resource_start(pPci_Dev, 0), pci_resource_len(pPci_Dev, 0));
	if (!csr_addr) 
	{
		DBGPRINT(RT_DEBUG_TRACE, "ioremap failed for device %s, region 0x%X @ 0x%lX\n",
			pPci_Dev->slot_name, (ULONG)pci_resource_len(pPci_Dev, 0), pci_resource_start(pPci_Dev, 0));
		goto err_out_free_mmio_region;
	}

	// Save CSR virtual address and irq to device structure
	net_dev->base_addr = csr_addr;
	pAd = net_dev->priv;
	pAd->CSRBaseAddress = csr_addr;
	pAd->net_dev = net_dev;

	// Set DMA master
	pci_set_master(pPci_Dev);

	// Read MAC address
	// 3. Read MAC address
	RTMP_IO_READ32(pAd, CSR3, &StaMacReg0.word);
	RTMP_IO_READ32(pAd, CSR4, &StaMacReg1.word);
	net_dev->dev_addr[0] = StaMacReg0.field.Byte0;
	net_dev->dev_addr[1] = StaMacReg0.field.Byte1;
	net_dev->dev_addr[2] = StaMacReg0.field.Byte2;
	net_dev->dev_addr[3] = StaMacReg0.field.Byte3;
	net_dev->dev_addr[4] = StaMacReg1.field.Byte4;
	net_dev->dev_addr[5] = StaMacReg1.field.Byte5;
	
	pAd->chip_id = chip_id;
	pAd->pPci_Dev = pPci_Dev;

	// The chip-specific entries in the device structure.
	net_dev->open = RT2400_open;
	net_dev->hard_start_xmit = RTMPSendPackets;
	net_dev->stop = RT2400_close;
	net_dev->get_stats = RT2400_get_ether_stats;

#if WIRELESS_EXT >= 12
	net_dev->get_wireless_stats = RT2400_get_wireless_stats;
#endif

	net_dev->set_multicast_list = RT2400_set_rx_mode;
	net_dev->do_ioctl = RT2400_ioctl;

	{// find available 
		int		i=0;
		char	slot_name[IFNAMSIZ];
		struct net_device	*device;

		for (i = 0; i < 8; i++)
		{
			sprintf(slot_name, "ra%d", i);
			
			for (device = dev_base; device != NULL; device = device->next)
			{
				if (strncmp(device->name, slot_name, 4) == 0)
				{
					break;
				}
			}
			if(device == NULL)	break;
		}

		sprintf(net_dev->name, "ra%d", i);
	}

	// Register this device
	Status = register_netdev(net_dev);
	if (Status)
		goto err_out_unmap;

	DBGPRINT(RT_DEBUG_TRACE, "%s: at 0x%lx, VA 0x%1x, IRQ %d. \n", 
		net_dev->name, pci_resource_start(pPci_Dev, 0), (ULONG)csr_addr, pPci_Dev->irq);

	// Set driver data
	pci_set_drvdata(pPci_Dev, net_dev);

	return 0;

err_out_unmap:
	iounmap((void *)csr_addr);
err_out_free_mmio_region:
	release_mem_region(pci_resource_start(pPci_Dev, 0), pci_resource_len(pPci_Dev, 0));
err_out_free_res:
	pci_release_regions(pPci_Dev);
err_out_free_netdev:
	kfree (net_dev);
err_out:
	return -ENODEV;
}


INT RT2400_open(
	IN	struct net_device *net_dev)
{
	PRTMP_ADAPTER	pAd = net_dev->priv;
	CSR3_STRUC	StaMacReg0;
	CSR4_STRUC	StaMacReg1;
	INT		status;

	// 1. allocate DMA buffers
	status = RTMPAllocDMAMemory(pAd);
	if (status)
		return status;
	
	// 2. request interrupt
	//
	// Disable interrupts here which is as soon as possible
	// This statement should never be true. We might consider to remove it later
	//
	if (RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_INTERRUPT_ACTIVE))
	{
		NICDisableInterrupt(pAd);
	}

	status = request_irq(pAd->pPci_Dev->irq, &RTMPIsr, SA_SHIRQ, net_dev->name, net_dev);
	if (status)
		return status;
	RTMP_SET_FLAG(pAd, fRTMP_ADAPTER_INTERRUPT_IN_USE);

	// 3. Read MAC address
	NICReadAdapterInfo(pAd);

	DBGPRINT(RT_DEBUG_TRACE, "%s: RT2400_open() irq %d. MAC = %02x:%02x:%02x:%02x:%02x:%02x \n",
		net_dev->name, pAd->pPci_Dev->irq, pAd->CurrentAddress[0], pAd->CurrentAddress[1], pAd->CurrentAddress[2],
		pAd->CurrentAddress[3], pAd->CurrentAddress[4], pAd->CurrentAddress[5]);

	NICInitTransmit(pAd);

	// manufacture default
 	PortCfgInit(pAd);

	// Read RaConfig profile parameters	 
	RTMPReadParametersFromFile(pAd);
	
	// initialize MLME
	status = MlmeInit(pAd);
	
	if(status != NDIS_STATUS_SUCCESS) {
	    return status;
	}

	pAd->MediaState = NdisMediaStateDisconnected;
	RTMP_SET_FLAG(pAd, fRTMP_ADAPTER_MEDIA_STATE_CHANGE);
	
	// Initialize Asics
	NICInitializeAdapter(pAd);

	NICReadEEPROMParameters(pAd);

	NICInitAsicFromEEPROM(pAd);
		
	// Start net interface tx /rx
	netif_start_queue(net_dev);

	// Set the timer to check for link beat.
	init_timer(&pAd->timer);
	pAd->timer.expires = jiffies + DEBUG_TASK_DELAY;
	pAd->timer.data = (unsigned long)net_dev;
	pAd->timer.function = &RT2400_timer;				// timer handler
	add_timer(&pAd->timer);

	netif_carrier_on(net_dev);

	// Enable interrupt
	RTMP_IO_WRITE32(pAd, CSR7, 0xffffffff); // clear garbage interrupts
	NICEnableInterrupt(pAd);

	return 0;
}

VOID RT2400_timer(
	IN	unsigned long data)
{
	struct net_device	*net_dev = (struct net_device *)data;
	RTMP_ADAPTER		*pAd = net_dev->priv;

//	NICCheckForHang(pAd);
	
	pAd->timer.expires = jiffies + DEBUG_TASK_DELAY;
	add_timer(&pAd->timer);
}

/*
	========================================================================
	
	Routine Description:
		hard_start_xmit handler

	Arguments:
		skb				point to sk_buf which upper layer transmit
		net_dev			point to net_dev
	Return Value:
		None

	Note:
	
	========================================================================
*/
INT	RTMPSendPackets(
	IN	struct sk_buff *skb,
	IN	struct net_device *net_dev)
{
	RTMP_ADAPTER	*pAd = net_dev->priv;
	NDIS_STATUS		Status = NDIS_STATUS_SUCCESS;
	
	if (!INFRA_ON(pAd) && !ADHOC_ON(pAd))
	{
		// Drop send request since there are no physical connection yet
		// Check the association status for infrastructure mode
		// And Mibss for Ad-hoc mode setup
		dev_kfree_skb_irq(skb);
	}
	else
	{
		if (skb->len > TX_BUFFER_SIZE)
		{
			// packet too long, drop it
			dev_kfree_skb_irq(skb);
			pAd->stats.tx_dropped++;
		}

		net_dev->trans_start = jiffies;
		Status = RTMPSendPacket(net_dev, skb);
		
		if (Status != NDIS_STATUS_SUCCESS)
		{
			//
			// Errors before enqueue stage
			//
			dev_kfree_skb_irq(skb);			
		}
	}

	// Dequeue one frame from SendTxWait queue and process it
	// There are two place calling dequeue for TX ring.
	// 1. Here, right after queueing the frame.
	// 2. At the end of TxRingTxDone service routine.
	if ((!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_BSS_SCAN_IN_PROGRESS)) && 
		(!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_RADIO_OFF)) &&
		(!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_RESET_IN_PROGRESS)))
	{
		RTMPDeQueuePacket(pAd, &pAd->SendTxWaitQueue);
	}

	return 0;
}

/*
	========================================================================
	
	Routine Description:
		Interrupt handler
		
	Arguments:
		irq							interrupt line
		dev_instance				Pointer to net_device
		rgs							store process's context before entering ISR, 
									this parameter is just for debug purpose.

	Return Value:
		VOID

	Note:

		
	========================================================================
*/
VOID RTMPIsr(
	IN	INT		irq, 
	IN	VOID		*dev_instance, 
	IN	struct pt_regs	*rgs)
{
	struct net_device	*net_dev = dev_instance;
	PRTMP_ADAPTER		pAd = net_dev->priv;
	INTSRC_STRUC		IntSource;

	// 1. Disable interrupt
	if (RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_INTERRUPT_IN_USE) && RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_INTERRUPT_ACTIVE))
	{
	        NICDisableInterrupt(pAd);
	}
	
	// 2. Get the interrupt sources & saved to local variable
	RTMP_IO_READ32(pAd, CSR7, &IntSource.word);
	RTMP_IO_WRITE32(pAd, CSR7, IntSource.word);

	//
	// Handle interrupt, walk through all bits
	// Should start from highest priority interrupt
	// The priority can be adjust by altering processing if statement
	//
	// If required spinlock, each interrupt service routine has to acquire
	// and release itself.
	//
	if (IntSource.field.TbcnExpire)
	{
		DBGPRINT(RT_DEBUG_INFO, "====> RTMPHandleTbcnInterrupt\n");
		RTMPHandleTbcnInterrupt(pAd);
	}

	if (IntSource.field.TwakeExpire)
	{
		DBGPRINT(RT_DEBUG_TRACE, "====> RTMPHandleTwakeupInterrupt\n");
		RTMPHandleTwakeupInterrupt(pAd);
	}

	if (IntSource.field.TatimwExpire)
	{
		DBGPRINT(RT_DEBUG_TRACE, "====> RTMPHandleTatimInterrupt\n");
		// RTMPHandleTatimInterrupt(pAd);
	}

	if (IntSource.field.TxRingTxDone)
	{
		DBGPRINT(RT_DEBUG_INFO, "====> RTMPHandleTxRingTxDoneInterrupt\n");
		RTMPHandleTxRingTxDoneInterrupt(pAd);
	}

	if (IntSource.field.AtimRingTxDone)
	{
		DBGPRINT(RT_DEBUG_INFO, "====> RTMPHandleAtimRingTxDoneInterrupt\n");
		RTMPHandleAtimRingTxDoneInterrupt(pAd);
	}

	if (IntSource.field.PrioRingTxDone)
	{
		DBGPRINT(RT_DEBUG_INFO, "====> RTMPHandlePrioRingTxDoneInterrupt\n");
		RTMPHandlePrioRingTxDoneInterrupt(pAd);
	}

	if (IntSource.field.RxDone)
	{
		DBGPRINT(RT_DEBUG_INFO, "====> RTMPHandleRxDoneInterrupt\n");
		RTMPHandleRxDoneInterrupt(net_dev);
	}

	// Do nothing if Reset in progress
	if (RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_RESET_IN_PROGRESS))
	{
		return;
	}

	// 3. Re-enable the interrupt, disabled earlier
	NICEnableInterrupt(pAd);
}

#if WIRELESS_EXT >= 12
/*
	========================================================================
	
	Routine Description:
		get wireless statistics

	Arguments:
		net_dev						Pointer to net_device

	Return Value:
		struct iw_statistics

	Note:
		This function will be called when query /proc
		
	========================================================================
*/
struct iw_statistics *RT2400_get_wireless_stats(
	IN	struct net_device *net_dev)
{
	RTMP_ADAPTER *pAd = net_dev->priv;

	DBGPRINT(RT_DEBUG_TRACE, "RT2400_get_wireless_stats --->\n");
	
	// TODO: All elements are zero before be implemented
	
	pAd->iw_stats.status = 0;						// Status - device dependent for now
	
	pAd->iw_stats.qual.qual = pAd->Mlme.RoamCqi;			// link quality (%retries, SNR, %missed beacons or better...)
	pAd->iw_stats.qual.level = pAd->PortCfg.LastRssi;	// signal level (dBm)
		
	pAd->iw_stats.qual.noise = 0;			// noise level (dBm)
	pAd->iw_stats.qual.updated = 1;		// Flags to know if updated

	pAd->iw_stats.discard.nwid = 0;		// Rx : Wrong nwid/essid
	pAd->iw_stats.miss.beacon = 0;		// Missed beacons/superframe

	// pAd->iw_stats.discard.code, discard.fragment, discard.retries, discard.misc has counted in other place
	
	return &pAd->iw_stats;
}
#endif

/*
	========================================================================
	
	Routine Description:
		return ethernet statistics counter

	Arguments:
		net_dev						Pointer to net_device

	Return Value:
		net_device_stats*

	Note:
		
		
	========================================================================
*/
struct net_device_stats *RT2400_get_ether_stats(
	IN	struct net_device *net_dev)
{
	RTMP_ADAPTER *pAd = net_dev->priv;

	DBGPRINT(RT_DEBUG_INFO, "RT2400_get_ether_stats --->\n");

	pAd->stats.rx_packets = pAd->WlanCounters.ReceivedFragmentCount;		// total packets received
	pAd->stats.tx_packets = pAd->WlanCounters.TransmittedFragmentCount;		// total packets transmitted
	
	pAd->stats.rx_bytes= pAd->RalinkCounters.ReceivedByteCount;				// total bytes received
	pAd->stats.tx_bytes = pAd->RalinkCounters.TransmittedByteCount;			// total bytes transmitted
	
	pAd->stats.rx_errors = pAd->Counters.RxErrors;							// bad packets received
	pAd->stats.tx_errors = pAd->Counters.TxErrors;							// packet transmit problems

	pAd->stats.rx_dropped = pAd->Counters.RxNoBuffer;						// no space in linux buffers
	pAd->stats.tx_dropped = pAd->WlanCounters.FailedCount;					// no space available in linux

	pAd->stats.multicast = pAd->WlanCounters.MulticastReceivedFrameCount;	// multicast packets received
	pAd->stats.collisions = pAd->Counters.OneCollision + pAd->Counters.MoreCollisions;	// Collision packets

	pAd->stats.rx_length_errors = 0;
	pAd->stats.rx_over_errors = pAd->Counters.RxNoBuffer;					// receiver ring buff overflow
	pAd->stats.rx_crc_errors = 0;//pAd->WlanCounters.FCSErrorCount;		// recved pkt with crc error
	pAd->stats.rx_frame_errors = pAd->Counters.RcvAlignmentErrors;			// recv'd frame alignment error
	pAd->stats.rx_fifo_errors = pAd->Counters.RxNoBuffer;					// recv'r fifo overrun
	pAd->stats.rx_missed_errors = 0;											// receiver missed packet
	
	// detailed tx_errors
	pAd->stats.tx_aborted_errors = 0;
	pAd->stats.tx_carrier_errors = 0;
	pAd->stats.tx_fifo_errors = 0;
	pAd->stats.tx_heartbeat_errors = 0;
	pAd->stats.tx_window_errors = 0;
	
	// for cslip etc
	pAd->stats.rx_compressed = 0;
	pAd->stats.tx_compressed = 0;
	
	return &pAd->stats;
}

/*
	========================================================================
	
	Routine Description:
		Set to filter multicast list

	Arguments:
		net_dev						Pointer to net_device

	Return Value:
		VOID

	Note:
		
		
	========================================================================
*/
VOID RT2400_set_rx_mode(
	IN	struct net_device *net_dev)
{
	RTMP_ADAPTER *pAd = net_dev->priv;
	// TODO: set_multicast_list
}

//
// Close driver function
//
INT	RT2400_close(
	IN	struct net_device *net_dev)
{
	RTMP_ADAPTER	*pAd = net_dev->priv;
	LONG			ioaddr = net_dev->base_addr;

	DBGPRINT(RT_DEBUG_TRACE, "%s: ===> RT2400_close\n", net_dev->name);

	// Disable interrupt first
	if (RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_INTERRUPT_ACTIVE))
	{
		NICDisableInterrupt(pAd);
	}

	// Stop Mlme state machine
	MlmeHalt(pAd);
	del_timer_sync(&pAd->PortCfg.RfTuningTimer);

	netif_stop_queue(net_dev);
	netif_carrier_off(net_dev);
	
	// Shut down monitor timer task
	del_timer_sync(&pAd->timer);

	// Disable Rx, register value supposed will remain after reset
	NICIssueReset(pAd);
	
	// Free IRQ
	if (RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_INTERRUPT_IN_USE))
	{
		// Deregister interrupt function
		free_irq(net_dev->irq, net_dev);
		RTMP_CLEAR_FLAG(pAd, fRTMP_ADAPTER_INTERRUPT_IN_USE);
	}

	// Free Ring buffers
	RTMPFreeDMAMemory(pAd);

	return 0;
}

//
// Remove driver function
//
static VOID __devexit RT2400_remove_one(
	IN	struct pci_dev	*pPci_Dev)
{
	struct net_device	*net_dev = pci_get_drvdata(pPci_Dev);
	RTMP_ADAPTER		*pAd = net_dev->priv;

	// Unregister network device
	unregister_netdev(net_dev);

	// Unmap CSR base address
	iounmap((char *)(net_dev->base_addr));
	
	// release memory region
	release_mem_region(pci_resource_start(pPci_Dev, 0), pci_resource_len(pPci_Dev, 0));	

	// Free pre-allocated net_device memory
	kfree(net_dev);
}

//
// Ralink PCI device table, include all supported chipsets
//
static struct pci_device_id rt2400_pci_tbl[] __devinitdata =
{
	{0x1814, 0x0101, PCI_ANY_ID, PCI_ANY_ID, 0, 0, RT2460A},
	{0,}			                    /* terminate list */
};
MODULE_DEVICE_TABLE(pci, rt2400_pci_tbl);

//
// Our PCI driver structure
//
static struct pci_driver rt2400_driver =
{
	name:		"rt2400",
	id_table:	rt2400_pci_tbl,
	probe:		RT2400_init_one,
#if LINUX_VERSION_CODE >= 0x20412		// Red Hat 7.3
	remove:		__devexit_p(RT2400_remove_one),
#else
	remove:		__devexit(RT2400_remove_one),
#endif
};

// =======================================================================
// LOAD / UNLOAD sections
// =======================================================================
//
// Driver module load function
//
static INT __init rt2400_init_module(VOID)
{
#ifdef EXPORT_NO_SYMBOLS
	EXPORT_NO_SYMBOLS;		// must use in init_module to avoid output symbol to /proc/ksyms
#endif
	return pci_module_init(&rt2400_driver);
}

//
// Driver module unload function
//
static VOID __exit rt2400_cleanup_module(VOID)
{
	pci_unregister_driver(&rt2400_driver);
}

module_init(rt2400_init_module);
module_exit(rt2400_cleanup_module);
